import Std
import Mathlib.Tactic.LibrarySearch

inductive Ty : Type :=
  | unit
  | arrow (A B : Ty)
deriving Repr, DecidableEq
open Ty

infixr:30 " ==> " => arrow

inductive Expr : Type :=
  | ann (e : Expr) (τ : Ty)
  | var (x : String)
  | abs (x : String) (e : Expr)
  | app (t₁ t₂ : Expr)
  | tt
deriving Repr, DecidableEq
open Expr

notation e " @@ " τ => ann e τ
notation "λ" x ", " e => abs x e
notation e₁ " ⋅ " e₂ => app e₁ e₂
notation "≪" x "≫" => var x

abbrev Context : Type := List (String × Ty)

inductive WellTyped : Context → Expr → Ty → Prop where
  | wt_var {Γ x τ} :
    Γ.lookup x = some τ →
    WellTyped Γ ≪x≫ τ
  | wt_ann {Γ e τ} :
    WellTyped Γ e τ →
    WellTyped Γ (e @@ τ) τ
  | wt_tt {Γ} :
    WellTyped Γ tt unit
  | wt_abs {Γ x τ₁ e τ₂} :
    WellTyped ((x, τ₁) :: Γ) e τ₂ →
    WellTyped Γ (λ x, e) (τ₁ ==> τ₂)
  | wt_app {Γ e₁ e₂ τ₁ τ₂} :
    WellTyped Γ e₁ (τ₁ ==> τ₂) →
    WellTyped Γ e₂ τ₁ →
    WellTyped Γ (e₁ ⋅ e₂) τ₂

notation Γ " ⊢ " e " : " τ => WellTyped Γ e τ

mutual
  def checkType : Context → Expr → Ty → Bool
    | _, tt, unit => true
    | Γ, (λ x, e), τ₁ ==> τ₂ => checkType ((x, τ₁) :: Γ) e τ₂
    | Γ, e, τ => match inferType Γ e with
      | some τ' => τ = τ'
      | none => false

  def inferType : Context → Expr → Option Ty
    | Γ, ≪x≫ => Γ.lookup x
    | Γ, e @@ τ => match checkType Γ e τ with
      | true => some τ
      | false => none
    | Γ, e₁ ⋅ e₂ => do
      let τ₁ ← inferType Γ e₁
      match τ₁ with
      | τa ==> τb => if checkType Γ e₂ τa then some τb else none
      | _ => none
    | _, tt => some unit
    | _, (λ _, _) => none
end

section
open WellTyped

mutual
  theorem inferType_sound :
    ∀ {Γ e τ},
    inferType Γ e = some τ →
    Γ ⊢ e : τ
    | Γ, tt, τ => by
      intro h; simp [checkType, inferType] at h
      simp [wt_tt, ←h]
    | Γ, var x, τ => by
      intro h; simp [checkType, inferType] at h
      simp [wt_var, h]
    | Γ, ann e τ', τ => by
      intro h; simp [checkType, inferType] at h
      match hct : checkType Γ e τ' with
      | false => simp [hct] at h
      | true =>
        simp [hct] at h
        rw [h]
        rw [h] at hct
        apply wt_ann
        apply checkType_sound
        apply hct
    | Γ, app e₁ e₂, τ => by
      intro h; simp [checkType, inferType] at h
      let ⟨τ₁, τ₂, hτ⟩ := h
      let τa ==> τb := τ₁
      simp at hτ
      let ⟨hτ₁, hτ₂⟩ := hτ
      rw [←hτ₂]
      apply @wt_app Γ e₁ e₂ τa τb
      . simp [inferType_sound, *]
      . apply checkType_sound; simp [*]
    | Γ, abs x e, τ => by
      intro h; simp [checkType, inferType] at h

  theorem checkType_sound :
    ∀ {Γ e τ},
    checkType Γ e τ = true →
    Γ ⊢ e : τ
    | Γ, tt, τ => by
      intro h
      cases τ with
      | unit => simp [wt_tt]
      | _ => simp [checkType, inferType] at h
    | Γ, abs x e, τ => by
      intro h
      match τ with
      | τ₁ ==> τ₂ =>
        simp [checkType, inferType] at h
        apply wt_abs
        apply checkType_sound; simp [*]
      | unit => simp [checkType, inferType] at h
    | Γ, var x, τ => by
      intro h
      simp [checkType] at h
      match hinfer : (inferType Γ (var x)) with
      | some τ' =>
        simp [hinfer] at h
        rw [h]
        apply inferType_sound; simp [*]
      | none =>
        rw [hinfer] at h
        simp [hinfer] at h
    | Γ, ann e τ', τ => by
      intro h
      simp [checkType] at h
      match hinfer : (inferType Γ (ann e τ')) with
      | some τ' =>
        simp [hinfer] at h
        rw [h]
        apply inferType_sound; simp [*]
      | none =>
        rw [hinfer] at h
        simp [hinfer] at h
    | Γ, app e₁ e₂, τ => by
      intro h
      simp [checkType] at h
      match hinfer : (inferType Γ (app e₁ e₂)) with
      | some τ' =>
        simp [hinfer] at h
        rw [h]
        apply inferType_sound; simp [*]
      | none =>
        rw [hinfer] at h
        simp [hinfer] at h
end
